"""Set client_add IP that we can trust."""

def includeme(config):
    config.add_tween('pyramid_heroku.client_addr.ClientAddr')


class ClientAddr(object):
    """Set request.client_addr to an IP we can trust.

    Sets correct IP to request.client_addr when running on Heroku with
    gunicorn. Should execute before other tweens which depend on it. This can
    be achieved with :ref:`Explicit tween ordering
    <pyramid:explicit_tween_ordering>` or by passing 'over' and 'under'
    arguments to :py:meth:`config.Configurator.add_tween` like this

    .. code-block:: python

        config.add_tween(
            'pyramid_heroku.client_addr.ClientAddr',
            under=pyramid.tweens.INGRESS,
        )

    The tween checks if ``X-Forwarded-For`` is present and if it is, takes
    the last IP in the ``X-Forwarded-For`` list and makes it the only IP
    in the list. This effectively causes pyramid to return this real IP for
    ``request.client_addr``. Read rationale behind why we do this on
    https://github.com/niteoweb/heroku_ips.
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        if request.headers.get('X-Forwarded-For'):
            request.headers['X-Forwarded-For'] = \
                request.headers['X-Forwarded-For'].split(',')[-1].strip()
        return self.handler(request)



"""Graceful DB migrations."""

from requests import Session
from time import sleep

import argparse
import os
import shlex
import subprocess
import sys


class Heroku(object):

    api_endpoint = 'https://api.heroku.com'

    def __init__(self, app_name, ini_file, app_section):
        # type: (str, str)-> ()
        """
        :param app_name: Name of Heroku app or id.
        :param app_section: Config section of app that needs migration.
        """
        self._formation = None
        self.app_name = app_name
        self.ini_file = ini_file
        self.app_section = app_section

        headers = {
            f'Authorization': 'Bearer {self.auth_key}',
            'Accept': 'application/vnd.heroku+json; version=3',
            'Content-Type': 'application/json',
        }
        self.session = Session()
        self.session.headers.update(headers)

    @property
    def auth_key(self):
        # type: () -> str
        """Heroku API secret.

        https://devcenter.heroku.com/articles/platform-api-quickstart#authentication.
        """
        return os.environ.get('MIGRATE_API_SECRET_HEROKU')

    def scale_down(self):
        """Scale all app workers to 0."""
        updates = [dict(type=t, quantity=0) for t in self.formation.keys()]
        res = self.session.patch(
            f'{self.api_endpoint}/apps/{self.app_name}/formation',
            json=dict(updates=updates),
        )
        self.parse_response(res)
        print('Scaled down to:')
        for x in res.json():
            print(f'{x["type"]}={x["quantity"]}')

    def scale_up(self):
        """Scale app back to initial state."""
        updates = [dict(type=t, quantity=s) for t, s in self.formation.items()]
        res = self.session.patch(
            f'{self.api_endpoint}/apps/{self.app_name}/formation',
            json=dict(updates=updates),
        )
        self.parse_response(res)
        print('Scaled up to:')
        for x in res.json():
            print(f'{x["type"]}={x["quantity"]}')

    @property
    def formation(self):
        """Get current app status and configuration.

        :return: Heroku app status as dict.
        """
        if not self._formation:
            res = self.session.get(
                f'{self.api_endpoint}/apps/{self.app_name}/formation')
            self.parse_response(res)
            self._formation = {x['type']: x['quantity'] for x in res.json()}
        return self._formation

    def shell(self, cmd):
        # type: (str) -> str
        """
        Run shell command.

        :param cmd: shell command to run
        :return: stdout of command
        """
        return subprocess.check_output(shlex.split(cmd)).decode()

    def needs_migrate(self):
        """
        Checks current status of alembic branches.

        :return: True if current alembic branch is not up to date.
        """
        cmd = self.shell(f'alembic -c {self.ini_file}'
                         f' -n {self.app_section} current')
        print(cmd)
        return 'head' not in cmd

    def alembic(self):
        """
        Run alembic migrations.

        :return: alembic stdout
        """
        return self.shell(f'alembic -c {self.ini_file}'
                          f' -n {self.app_section} upgrade head')

    def set_maintenance(self, state):
        # type: (bool) -> ()
        res = self.session.patch(
            f'{self.api_endpoint}/apps/{self.app_name}',
            json=dict(maintenance=state),
        )
        if self.parse_response(res):
            print('Maintenance {}'.format('enabled' if state else 'disabled'))

    def parse_response(self, res):
        # type: (requests.Response) -> bool
        """
        Parses Heroku API response.

        :param res: requests object
        :return: true if request succeeded
        """
        if res.status_code != 200:
            print(res.json())
        res.raise_for_status()
        return True

    def migrate(self):
        """Run database migration if needed."""
        print(self.app_name)
        print(self.app_section)
        print(self.ini_file)

        if self.needs_migrate():
            self.set_maintenance(True)
            self.scale_down()
            sleep(30)
            print(self.alembic())
            self.scale_up()
            sleep(30)
            self.set_maintenance(False)
        else:
            print('Database migration is not needed')


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=(
            'usage: migrate.py [-h] app_name [ini_file] [app_section]'
            '\nexample: python -m pyramid_heroku.migrate my_app etc/production.ini app:main'),
    )
    parser.add_argument('app_name', help='Heroku app name')
    parser.add_argument(
        'ini_file',
        nargs='?',
        default='etc/production.ini',
        help='Path to Pyramid configuration file ')
    parser.add_argument(
        'app_section',
        nargs='?',
        default='app:main',
        help='App section name in ini configuration file')

    options = parser.parse_args()

    Heroku(options.app_name, options.ini_file, options.app_section).migrate()


if __name__ == '__main__':
    main()



"""Various utilities."""

from ast import literal_eval

import os
import sys
import typing as t


def safe_eval(text: str) -> str:
    """Safely evaluate `text` argument.

    `text` can be evaluated to string, number,
    tuple, list, dict, boolean, and None.
    Code `text[0].upper + text[1:]` is for lower case
    'false' or 'true' strings, it should not break anything.
    """
    if not isinstance(text, str):
        raise ValueError(f"Expected a string, got {type(text)} instead.")

    try:
        if len(text) == 0:
            return None
        else:
            return literal_eval(text[0].upper() + text[1:])
    except (ValueError, SyntaxError):
        return text


def expandvars_dict(settings: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """Expand all environment variables in a settings dictionary.

    Env vars are expanded twice, so you can have (1-level) nesting. Example:

    $ export FOO='foo'
    $ export BAR='${FOO}bar'
    $ python
    >>> from pyramid_heroku import expandvars_dict
    >>> expandvars_dict({'my_var': '${BAR}'})
    {'my_var': 'foobar'}
    """
    return {
        key: safe_eval(os.path.expandvars(os.path.expandvars(value)))
        for (key, value) in settings.items()
    }



import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
readme = open(os.path.join(here, 'README.rst')).read()
changes = open(os.path.join(here, 'CHANGES.rst')).read()


requires = [
    'pyramid>=1.7',
    'requests',
]


setup(
    name='pyramid_heroku',
    version='0.3',
    description='A bunch of helpers for successfully running Pyramid on Heroku.',
    long_description=readme + '\n' + changes,
    classifiers=[
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Framework :: Pylons",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    packages=['pyramid_heroku'],
    install_requires=requires,
    author='Niteo',
    author_email='info@niteo.co',
    license='BSD',
    url='https://github.com/niteoweb/pyramid_heroku',
    keywords='pyramid heroku pylons web',
    test_suite='pyramid_heroku',
)
